from google.cloud import storage
from google.cloud.storage.notification import JSON_API_V1_PAYLOAD_FORMAT


class EmsStorageClient:
    def __init__(self, project_id: str):
        self.project_id = project_id
        self.__client = storage.Client(project_id)

    def download_lines(self, bucket_name: str, blob_name: str, num_lines: int, chunk_size: int = 1024 * 1024) -> list:
        bucket = self.__client.bucket(bucket_name)
        blob = bucket.blob(blob_name)
        buffer = blob.download_as_string(start=0, end=chunk_size)
        lines = buffer.decode("utf-8", errors="ignore").split("\n")
        if len(lines) < num_lines:
            raise NotImplementedError("First chunk does not contain enough lines, increase chunk size")
        return lines[0:num_lines]

    def upload_from_string(self, bucket_name: str, blob_name: str, content: str):
        bucket = self.__client.bucket(bucket_name)
        blob = bucket.blob(blob_name)
        blob.upload_from_string(content)

    def download_content_as_string(self, bucket_name: str, blob_name: str):
        bucket = self.__client.bucket(bucket_name)
        blob = bucket.blob(blob_name)
        return blob.download_as_string().decode("utf-8")

    def create_bucket_if_not_exists(self, bucket_name: str, project=None, location=None):
        bucket = self.__client.bucket(bucket_name)
        if not bucket.exists():
            bucket.create(project=project, location=location)

    def delete_bucket_if_exists(self, bucket_name: str, force: bool = False):
        bucket = self.__client.bucket(bucket_name)
        if bucket.exists():
            bucket.delete(force=force)

    def create_notification_if_not_exists(self, topic_name: str, bucket_name: str):
        bucket = self.__client.bucket(bucket_name)

        if not self.is_notification_exists(bucket, topic_name):
            bucket.notification(topic_name, payload_format=JSON_API_V1_PAYLOAD_FORMAT).create()

    def delete_notification_if_exists(self, topic_name: str, bucket_name: str):
        bucket = self.__client.bucket(bucket_name)

        for notification in bucket.list_notifications():
            if notification.topic_name == topic_name:
                notification.delete()

    def is_blob_empty(self, bucket_name: str, blob_name: str):
        return self.__client.get_bucket(bucket_name).get_blob(blob_name).size == 0

    @staticmethod
    def is_notification_exists(bucket, topic_name):
        for notification in bucket.list_notifications():
            if notification.topic_name == topic_name:
                return True
        return False

    def delete_blob(self, bucket_name: str, blob_name: str):
        bucket = self.__client.bucket(bucket_name)
        blob = bucket.blob(blob_name)
        blob.delete()



import time
from unittest import TestCase

from google.api_core.exceptions import NotFound, AlreadyExists
from google.cloud.pubsub_v1 import PublisherClient, SubscriberClient

from pubsub.ems_publisher_client import EmsPublisherClient
from pubsub.ems_subscriber_client import EmsSubscriberClient
from tests.integration import GCP_PROJECT_ID


class ItEmsSubscriberClient(TestCase):

    def setUp(self):
        self.__ems_publisher_client = EmsPublisherClient()
        self.__ems_subscriber_client = EmsSubscriberClient()
        self.__publisher_client = PublisherClient()
        self.__subscriber_client = SubscriberClient()

    def test_create_subscription_if_not_exists(self):
        topic_name = self.__generate_test_name("topic")
        subscription_name = self.__generate_test_name("subscription")

        self.__ems_publisher_client.topic_create_if_not_exists(GCP_PROJECT_ID, topic_name)
        self.__ems_subscriber_client.create_subscription_if_not_exists(GCP_PROJECT_ID, topic_name, subscription_name)
        try:
            self.__ems_subscriber_client.create_subscription_if_not_exists(GCP_PROJECT_ID, topic_name, subscription_name)
        except AlreadyExists:
            self.fail("create_subscription_if_not_exists raised AlreadyExists error")

        topic_path = self.__publisher_client.api.topic_path(GCP_PROJECT_ID, topic_name)
        subscriptions = list(self.__publisher_client.api.list_topic_subscriptions(request={"topic": topic_path}))

        expected_subscriptions = ["projects/" + GCP_PROJECT_ID + "/subscriptions/" + subscription_name]
        self.assertEqual(expected_subscriptions, subscriptions)

        self.__delete_subscription(subscription_name)
        self.__delete_topic(topic_name)

    def test_delete_subscription_if_exists(self):
        topic_name = self.__generate_test_name("topic")
        subscription_name = self.__generate_test_name("subscription")

        self.__ems_publisher_client.topic_create_if_not_exists(GCP_PROJECT_ID, topic_name)
        self.__ems_subscriber_client.create_subscription_if_not_exists(GCP_PROJECT_ID, topic_name, subscription_name)
        self.__ems_subscriber_client.delete_subscription_if_exists(GCP_PROJECT_ID, subscription_name)
        try:
            self.__ems_subscriber_client.delete_subscription_if_exists(GCP_PROJECT_ID, subscription_name)
        except NotFound:
            self.fail("delete_subscription_if_exists raised NotFound error")

        subscription_path = self.__subscriber_client.api.subscription_path(GCP_PROJECT_ID, subscription_name)
        with self.assertRaises(NotFound):
            self.__subscriber_client.api.get_subscription(request={"subscription": subscription_path})

        self.__delete_topic(topic_name)

    def __delete_topic(self, topic_name):
        topic_path = self.__publisher_client.api.topic_path(GCP_PROJECT_ID, topic_name)
        self.__publisher_client.api.delete_topic(request={"topic": topic_path})

    def __delete_subscription(self, subscription_name: str):
        subscription_path = self.__subscriber_client.api.subscription_path(GCP_PROJECT_ID, subscription_name)
        self.__subscriber_client.api.delete_subscription(request={"subscription": subscription_path})

    @staticmethod
    def __generate_test_name(context: str):
        return "test_" + context + "_" + str(int(time.time()))



import logging
from concurrent.futures import Future

from google.api_core.exceptions import AlreadyExists, NotFound
from google.cloud.pubsub_v1 import PublisherClient, SubscriberClient

LOGGER = logging.getLogger(__name__)


class EmsPublisherClient:

    def __init__(self):
        self.__client = PublisherClient()

    def publish(self, topic: str, data: bytes, **attrs) -> Future:
        return self.__client.publish(topic=topic, data=data, **attrs)

    def topic_create_if_not_exists(self, project_id: str, topic_name: str):
        topic_path = self.__client.api.topic_path(project_id, topic_name)
        try:
            self.__client.api.create_topic(request={"name": topic_path})
            LOGGER.info("Topic %s created in project %s", topic_name, project_id)
        except AlreadyExists:
            LOGGER.info("Topic %s already exists in project %s", topic_name, project_id)

    def delete_topic_if_exists(self, project_id: str, topic_name: str):
        topic_path = self.__client.api.topic_path(project_id, topic_name)
        try:
            self.__client.api.delete_topic(request={"topic": topic_path})
            LOGGER.info("Topic %s deleted in project %s", topic_name, project_id)
        except NotFound:
            LOGGER.info("Topic %s not found in project %s", topic_name, project_id)

    @staticmethod
    def subscription_create(project_id: str, topic_name: str, subscription_name: str):
        subscriber = SubscriberClient()

        topic_path = subscriber.api.topic_path(project_id, topic_name)
        subscription_path = subscriber.api.subscription_path(project_id, subscription_name)

        subscriber.api.create_subscription(request={"name": subscription_path, "topic": topic_path})



from abc import ABC, abstractmethod
from enum import Enum


class EmsWriteDisposition(Enum):
    WRITE_APPEND = "WRITE_APPEND"
    WRITE_TRUNCATE = "WRITE_TRUNCATE"
    WRITE_EMPTY = "WRITE_EMPTY"


class EmsCreateDisposition(Enum):
    CREATE_IF_NEEDED = "CREATE_IF_NEEDED"
    CREATE_NEVER = "CREATE_NEVER"


class EmsJobPriority(Enum):
    INTERACTIVE = "INTERACTIVE"
    BATCH = "BATCH"


class EmsJobConfig(ABC):

    @abstractmethod
    def __init__(self,
                 destination_project_id: str = None,
                 destination_dataset: str = None,
                 destination_table: str = None,
                 create_disposition: EmsCreateDisposition = EmsCreateDisposition.CREATE_IF_NEEDED,
                 write_disposition: EmsWriteDisposition = EmsWriteDisposition.WRITE_APPEND,
                 table_definitions: dict = None,
                 labels: dict = {}) -> None:
        self.__destination_project_id = destination_project_id
        self.__destination_dataset = destination_dataset
        self.__create_disposition = create_disposition
        self.__write_disposition = write_disposition
        self.__destination_table = destination_table
        self.__table_definitions = table_definitions
        self.__labels = labels

    @property
    def destination_project_id(self):
        return self.__destination_project_id

    @property
    def destination_dataset(self):
        return self.__destination_dataset

    @property
    def destination_table(self):
        return self.__destination_table

    @property
    def create_disposition(self):
        return self.__create_disposition

    @property
    def write_disposition(self):
        return self.__write_disposition

    @property
    def table_definitions(self):
        return self.__table_definitions

    @property
    def labels(self):
        return self.__labels



import datetime
import logging
from typing import NamedTuple

from cloudsql.ems_cloudsql_client import EmsCloudsqlClient
from storage.ems_storage_client import EmsStorageClient

LOGGER = logging.getLogger(__name__)


# pylint: disable=C0103
class TempBucketDescriptor(NamedTuple):
    project_id: str
    name: str
    location: str


class EmsCloudsqlUtil:
    IMPORT_CSV_TIMEOUT = 600
    RELOAD_TABLE_TIMEOUT = 600
    CREATE_TMP_TABLE_TIMEOUT = 30

    def __init__(self,
                 cloud_sql_client: EmsCloudsqlClient,
                 storage_client: EmsStorageClient,
                 temp_bucket: TempBucketDescriptor):
        self.__temp_bucket = temp_bucket
        self.__storage_client = storage_client
        self.__cloud_sql_client = cloud_sql_client

    def run_sql(self, database: str, sql_query: str, import_user: str, timeout_seconds: float = 30) -> None:
        suffix = str(int(datetime.datetime.utcnow().timestamp()))
        blob_name = f"sql_query_{suffix}"
        LOGGER.info("Running SQL %s in database %s", sql_query, database)
        self.__storage_client.create_bucket_if_not_exists(bucket_name=self.__temp_bucket.name,
                                                          project=self.__temp_bucket.project_id,
                                                          location=self.__temp_bucket.location)
        self.__storage_client.upload_from_string(bucket_name=self.__temp_bucket.name,
                                                 blob_name=blob_name,
                                                 content=sql_query)
        self.__cloud_sql_client.import_sql_from_bucket(database=database,
                                                       source_sql_uri=f"gs://{self.__temp_bucket.name}/{blob_name}",
                                                       timeout_seconds=timeout_seconds,
                                                       import_user=import_user)
        self.__storage_client.delete_blob(bucket_name=self.__temp_bucket.name,
                                          blob_name=blob_name)



from unittest import TestCase
from unittest.mock import patch

from googleapiclient import discovery

from cloudsql.ems_cloudsql_client import EmsCloudsqlClient


class TestEmsCloudsqlClient(TestCase):

    # pylint:disable=unused-argument
    @patch("cloudsql.ems_cloudsql_client.discovery")
    def test_properties(self, discovery_module_patch: discovery):
        client = EmsCloudsqlClient("some-project-id", "some-instance-id")

        self.assertEqual(client.project_id, "some-project-id")
        self.assertEqual(client.instance_id, "some-instance-id")



from datetime import datetime
from abc import ABC
from typing import Union

from bigquery.job.ems_job_state import EmsJobState


class EmsJob(ABC):

    def __init__(self,
                 job_id: str,
                 state: EmsJobState,
                 error_result: Union[dict, None],
                 created: datetime = None):
        self.__job_id = job_id
        self.__state = state
        self.__error_result = error_result
        self.__created = created

    @property
    def state(self) -> EmsJobState:
        return self.__state

    @property
    def created(self) -> datetime:
        return self.__created

    @property
    def job_id(self) -> str:
        return self.__job_id

    @property
    def is_failed(self) -> bool:
        return self.__error_result is not None

    @property
    def error_result(self) -> Union[dict, None]:
        return self.__error_result



import logging
import re
from datetime import datetime
from typing import List, Union, Iterable

from google.api_core.exceptions import GoogleAPIError, NotFound, Conflict
from google.cloud import bigquery
from google.cloud.bigquery import QueryJobConfig, QueryJob, TableReference, DatasetReference, TimePartitioning, \
    LoadJobConfig, LoadJob, ExtractJobConfig, ExtractJob
from google.cloud.bigquery.schema import _parse_schema_resource, _build_schema_resource

from bigquery.ems_api_error import EmsApiError
from bigquery.job.config.ems_extract_job_config import EmsExtractJobConfig, Compression, DestinationFormat
from bigquery.job.config.ems_job_config import EmsJobPriority, EmsCreateDisposition, EmsWriteDisposition
from bigquery.job.config.ems_load_job_config import EmsLoadJobConfig
from bigquery.job.config.ems_query_job_config import EmsQueryJobConfig, EmsTimePartitioning, EmsTimePartitioningType
from bigquery.job.ems_extract_job import EmsExtractJob
from bigquery.job.ems_job import EmsJob
from bigquery.job.ems_job_state import EmsJobState
from bigquery.job.ems_load_job import EmsLoadJob
from bigquery.job.ems_query_job import EmsQueryJob

LOGGER = logging.getLogger(__name__)

RETRY = "-retry-"


class EmsBigqueryClient:
    def __init__(self, project_id: str, location: str = "EU"):
        self.__project_id = project_id
        self.__bigquery_client = bigquery.Client(project_id, location=location)
        self.__location = location

    @property
    def project_id(self) -> str:
        return self.__project_id

    @property
    def location(self) -> str:
        return self.__location

    def dataset_exists(self, dataset_id: str):
        try:
            self.__bigquery_client.get_dataset(dataset_id)
        except NotFound:
            return False
        return True

    def create_dataset_if_not_exists(self, dataset_id: str):
        dataset = self.__bigquery_client.dataset(dataset_id, self.__project_id)
        try:
            self.__bigquery_client.create_dataset(dataset)
            LOGGER.info("Dataset %s created in project %s", dataset_id, self.__project_id)
        except Conflict:
            LOGGER.info("Dataset %s already exists in project %s", dataset_id, self.__project_id)

    def delete_dataset_if_exists(self, dataset_id: str, delete_contents=False):
        try:
            self.__bigquery_client.delete_dataset(dataset_id, delete_contents)
            LOGGER.info("Dataset %s deleted in project %s", dataset_id, self.__project_id)
        except NotFound:
            LOGGER.info("Dataset %s not found in project %s", dataset_id, self.__project_id)

    def table_exists(self, table: str) -> bool:
        try:
            self.__bigquery_client.get_table(table=TableReference.from_string(table))
            return True
        except NotFound:
            return False

    def get_job_list(self, min_creation_time: datetime = None, max_creation_time: datetime = None, max_result: int = 20,
                     all_users: bool = True) -> Iterable:
        """
        Args:
            min_creation_time (datetime.datetime, optional):
                If set, only jobs created after or at this timestamp are returned.
                If the datetime has no time zone assumes UTC time.
            max_creation_time (datetime.datetime, optional):
                If set, only jobs created before or at this timestamp are returned.
                If the datetime has no time zone assumes UTC time.
            max_result (int, optional):
                Maximum number of jobs to return.
            all_users (bool):
                If true, retrieves jobs submitted by all users
        Yields:
            EmsQueryJob: the next job
        """
        for job in self.__bigquery_client.list_jobs(all_users=all_users,
                                                    max_results=max_result,
                                                    min_creation_time=min_creation_time,
                                                    max_creation_time=max_creation_time):
            ems_job = self.__convert_to_ems_job(job)

            if ems_job is not None:
                yield ems_job

    @staticmethod
    def __convert_to_ems_job(job):
        if isinstance(job, QueryJob):
            destination = job.destination
            table_id, dataset_id, project_id = \
                (destination.table_id, destination.dataset_id, destination.project) \
                    if destination is not None else (None, None, None)

            config = EmsQueryJobConfig(priority=EmsJobPriority[job.priority],
                                       destination_project_id=project_id,
                                       destination_dataset=dataset_id,
                                       destination_table=table_id,
                                       create_disposition=EmsBigqueryClient.__convert_to_ems_create_disposition(
                                           job.create_disposition),
                                       write_disposition=EmsBigqueryClient.__convert_to_ems_write_disposition(
                                           job.write_disposition),
                                       time_partitioning=EmsBigqueryClient.__convert_to_ems_time_partitioning(
                                           job.time_partitioning),
                                       labels=job.labels)
            return EmsQueryJob(job.job_id, job.query,
                               config,
                               EmsJobState(job.state),
                               job.error_result,
                               job.created)
        elif isinstance(job, LoadJob):
            destination = job.destination
            table_id, dataset_id, project_id = destination.table_id, destination.dataset_id, destination.project
            schema = {"fields": _build_schema_resource(job.schema)} if job.schema else []

            config = EmsLoadJobConfig(schema=schema,
                                      source_uri_template=job.source_uris[0] if job.source_uris else None,
                                      destination_project_id=project_id,
                                      destination_dataset=dataset_id,
                                      destination_table=table_id,
                                      create_disposition=EmsBigqueryClient.__convert_to_ems_create_disposition(
                                          job.create_disposition),
                                      write_disposition=EmsBigqueryClient.__convert_to_ems_write_disposition(
                                          job.write_disposition),
                                      labels=job.labels)

            return EmsLoadJob(job_id=job.job_id,
                              load_config=config,
                              state=EmsJobState(job.state),
                              error_result=None,
                              created=job.created)
        elif isinstance(job, ExtractJob):
            table = f'{job.source.project}.{job.source.dataset_id}.{job.source.table_id}'
            destination_uris = job.destination_uris
            job_config = EmsExtractJobConfig(
                compression=Compression(job.compression) if job.compression else Compression.NONE,
                destination_format=DestinationFormat(
                    job.destination_format) if job.destination_format else DestinationFormat.CSV,
                field_delimiter=job.field_delimiter,
                print_header=job.print_header,
                labels=job.labels)
            return EmsExtractJob(job_id=job.job_id,
                                 table=table,
                                 destination_uris=destination_uris,
                                 job_config=job_config,
                                 state=EmsJobState(job.state),
                                 error_result=job.error_result,
                                 created=job.created)
        else:
            LOGGER.warning(f"Unexpected job type for : {job.job_id}, with type class: {job.__class__}")
            return None

    @staticmethod
    def __convert_to_ems_create_disposition(disposition):
        if disposition is None:
            return None
        return EmsCreateDisposition(disposition)

    @staticmethod
    def __convert_to_ems_time_partitioning(partitioning: TimePartitioning) -> EmsTimePartitioning:
        if partitioning is None:
            return None
        return EmsTimePartitioning(type_=EmsTimePartitioningType(partitioning.type_),
                                   field=partitioning.field,
                                   expiration_ms=partitioning.expiration_ms,
                                   require_partition_filter=partitioning.require_partition_filter)

    @staticmethod
    def __convert_to_ems_write_disposition(disposition):
        if disposition is None:
            return None
        return EmsWriteDisposition(disposition)

    def get_jobs_with_prefix(self, job_prefix: str, min_creation_time: datetime, max_creation_time: datetime = None,
                             max_result: int = 20, all_users: bool = True) -> list:
        jobs = self.get_job_list(min_creation_time, max_creation_time, max_result, all_users=all_users)
        matched_jobs = filter(lambda x: job_prefix in x.job_id, jobs)
        return list(matched_jobs)

    def relaunch_failed_jobs(self, job_prefix: str, min_creation_time: datetime, max_creation_time: datetime = None,
                             max_attempts: int = 3, max_result: int = None, all_users: bool = True) -> list:
        def launch(job: Union[EmsQueryJob, EmsExtractJob]) -> str:
            prefix_with_retry = self.__decorate_id_with_retry(job.job_id, job_prefix, max_attempts)

            if isinstance(job, EmsQueryJob):
                return self.run_async_query(job.query, prefix_with_retry, job.query_config)
            elif isinstance(job, EmsExtractJob):
                return self.run_async_extract_job(prefix_with_retry, job.table, job.destination_uris, job.job_config)
            else:
                LOGGER.error(f"Unsupported job: {job}")

        jobs = self.get_jobs_with_prefix(job_prefix, min_creation_time, max_creation_time, max_result,
                                         all_users=all_users)
        failed_jobs = [x for x in jobs if x.is_failed]
        return [launch(job) for job in failed_jobs]

    def run_async_query(self,
                        query: str,
                        job_id_prefix: str = None,
                        ems_query_job_config: EmsQueryJobConfig = EmsQueryJobConfig(
                            priority=EmsJobPriority.INTERACTIVE)) -> str:
        return self.__execute_query_job(query=query,
                                        ems_query_job_config=ems_query_job_config,
                                        job_id_prefix=job_id_prefix).job_id

    def run_async_load_job(self, job_id_prefix: str, config: EmsLoadJobConfig) -> str:
        return self.__bigquery_client.load_table_from_uri(source_uris=config.source_uri_template,
                                                          destination=TableReference(
                                                              DatasetReference(config.destination_project_id,
                                                                               config.destination_dataset),
                                                              config.destination_table),
                                                          job_id_prefix=job_id_prefix,
                                                          location=self.__location,
                                                          job_config=self.__create_load_job_config(config)).job_id

    def run_async_extract_job(self, job_id_prefix: str, table: str, destination_uris: List[str],
                              job_config: EmsExtractJobConfig) -> str:

        extract_job_config = self.__create_extract_job_config(job_config)

        return self.__bigquery_client.extract_table(source=TableReference.from_string(table_id=table),
                                                    destination_uris=destination_uris,
                                                    job_id_prefix=job_id_prefix,
                                                    location=self.__location,
                                                    job_config=extract_job_config).job_id

    def run_sync_query(self,
                       query: str,
                       ems_query_job_config: EmsQueryJobConfig = EmsQueryJobConfig(priority=EmsJobPriority.INTERACTIVE),
                       job_id_prefix: str = None
                       ) -> Iterable:
        LOGGER.info("Sync query executed with priority: %s", ems_query_job_config.priority)
        try:
            return self.__get_mapped_iterator(
                self.__execute_query_job(
                    query=query,
                    ems_query_job_config=ems_query_job_config,
                    job_id_prefix=job_id_prefix
                ).result()
            )
        except GoogleAPIError as e:
            raise EmsApiError("Error caused while running query | {} |: {}!".format(query, e.args[0]))

    def wait_for_job_done(self, job_id: str, timeout_seconds: float) -> EmsJob:
        job = self.__bigquery_client.get_job(job_id, project=self.__project_id, location=self.__location)
        job.result(timeout=timeout_seconds)
        return self.__convert_to_ems_job(job)

    def __decorate_id_with_retry(self, job_id: str, job_prefix: str, retry_limit: int):
        retry_counter = 0
        if RETRY in job_id:
            retry_counter = self.__get_retry_counter(job_id, job_prefix)
        prefix_with_retry = job_prefix + RETRY + str(retry_counter + 1) + "-"

        if retry_counter >= retry_limit - 1:
            raise RetryLimitExceededError()
        return prefix_with_retry

    def __get_retry_counter(self, job_id, job_id_prefix):
        regex = job_id_prefix + RETRY + "([0-9]+)-.+"
        return int(re.search(regex, job_id).group(1))

    def __execute_query_job(self, query: str, ems_query_job_config: EmsQueryJobConfig, job_id_prefix=None) -> QueryJob:
        return self.__bigquery_client.query(query=query,
                                            job_config=(self.__create_job_config(ems_query_job_config)),
                                            job_id_prefix=job_id_prefix,
                                            location=self.__location,
                                            retry=bigquery.DEFAULT_RETRY.with_deadline(300))

    def __create_load_job_config(self, ems_load_job_config: EmsLoadJobConfig) -> LoadJobConfig:
        config = LoadJobConfig()
        config.labels = ems_load_job_config.labels
        config.create_disposition = ems_load_job_config.create_disposition.value
        config.write_disposition = ems_load_job_config.write_disposition.value
        config.schema = _parse_schema_resource(ems_load_job_config.schema)
        config.skip_leading_rows = ems_load_job_config.skip_leading_rows
        return config

    def __create_extract_job_config(self, ems_job_config: EmsExtractJobConfig) -> ExtractJobConfig:
        config = ExtractJobConfig()

        config.labels = ems_job_config.labels
        config.compression = ems_job_config.compression.value
        config.destination_format = ems_job_config.destination_format.value
        config.field_delimiter = ems_job_config.field_delimiter
        config.print_header = ems_job_config.print_header
        return config

    def __create_job_config(self, ems_query_job_config: EmsQueryJobConfig) -> QueryJobConfig:
        job_config = QueryJobConfig()
        job_config.priority = ems_query_job_config.priority.value
        job_config.use_legacy_sql = False
        job_config.use_query_cache = ems_query_job_config.use_query_cache
        job_config.labels = ems_query_job_config.labels
        if ems_query_job_config.destination_table is not None:
            job_config.time_partitioning = TimePartitioning("DAY")
            table_reference = TableReference(
                DatasetReference(ems_query_job_config.destination_project_id or self.__project_id,
                                 ems_query_job_config.destination_dataset),
                ems_query_job_config.destination_table)
            job_config.destination = table_reference
            job_config.write_disposition = ems_query_job_config.write_disposition.value
            job_config.create_disposition = ems_query_job_config.create_disposition.value
        partitioning = ems_query_job_config.time_partitioning
        if partitioning is not None:
            job_config.time_partitioning = TimePartitioning(partitioning.type.value,
                                                            partitioning.field,
                                                            partitioning.expiration_ms,
                                                            partitioning.require_partition_filter)
        if ems_query_job_config.table_definitions is not None:
            job_config.table_definitions = ems_query_job_config.table_definitions
        return job_config

    @staticmethod
    def __get_mapped_iterator(result: Iterable):
        for row in result:
            yield dict(list(row.items()))


class RetryLimitExceededError(Exception):
    pass



from unittest import TestCase

from pubsub.ems_message import EmsMessage


class TestEmsMessage(TestCase):

    @staticmethod
    def test_dataIsParsedAsJson():
        # pylint: disable=invalid-string-quote
        message = EmsMessage("ackId", b'{"a":"v"}', dict())

        assert {"a": "v"} == message.data_json

    @staticmethod
    def test_constructorWontThrowOnInvalidData():
        EmsMessage("ackId", b"Invalid json", dict())



import random
import time
from unittest import TestCase

from google.api_core.exceptions import NotFound
from google.cloud import storage

from pubsub.ems_publisher_client import EmsPublisherClient
from storage.ems_storage_client import EmsStorageClient
from tests.integration import GCP_PROJECT_ID

IT_TEST_BUCKET = "it_test_ems_gcp_toolkit"

TOOLKIT_CREATED_BUCKET = "it_test_ems_gcp_toolkit_created_bucket"

TOOLKIT_CREATED_TOPIC = "it_test_ems_gcp_toolkit_created_topic"


class ItEmsStorageClientTest(TestCase):

    def setUp(self):
        self.__ems_storage_client = EmsStorageClient(GCP_PROJECT_ID)
        self.__ems_publisher_client = EmsPublisherClient()

    @classmethod
    def setUpClass(cls):
        cls.__storage_client = storage.Client(GCP_PROJECT_ID)
        bucket = cls.__storage_client.bucket(IT_TEST_BUCKET)
        if not bucket.exists():
            bucket.location = "europe-west1"
            bucket.storage_class = "REGIONAL"
            bucket.create()

        cls.bucket = bucket

    @classmethod
    def tearDownClass(cls):
        bucket_name = TOOLKIT_CREATED_BUCKET
        bucket = cls.__storage_client.bucket(bucket_name)
        if bucket.exists():
            bucket.delete(force=True)

    def test_download_lines_downloadingSingleLine_returnsHeader(self):
        blob_name = "sample_test_with_header.csv"
        blob = self.bucket.blob(blob_name)
        num_cols = random.randint(1, 5)
        header = ",".join(["header"] * num_cols)
        blob.upload_from_string(f"{header}\nROW\n")

        gcs_header = self.__ems_storage_client.download_lines(self.bucket.name, blob_name, 1)

        self.assertEqual([header], gcs_header)

    def test_download_lines_downloadingMultipleLines_returnsRows(self):
        blob_name = "sample_multiline.txt"
        blob = self.bucket.blob(blob_name)
        blob.upload_from_string("line1\nline2\nline3\n")

        lines = self.__ems_storage_client.download_lines(self.bucket.name, blob_name, 2)

        self.assertEqual(["line1", "line2"], lines)

    def test_download_lines_ifReturnedLinesNotEqualsRequestedLines_raiseException(self):
        blob_name = "sample_big_multiline.txt"
        blob = self.bucket.blob(blob_name)
        lines = ["line"] * 10
        blob.upload_from_string("\n".join(lines))

        with self.assertRaises(NotImplementedError):
            self.__ems_storage_client.download_lines(self.bucket.name, blob_name, len(lines), 10)

    def test_upload_from_string(self):
        blob_name = "test_upload.txt"
        content = "Test data to upload"
        self.__ems_storage_client.upload_from_string(self.bucket.name, blob_name, content)

        blob = self.bucket.blob(blob_name)
        actual_content = blob.download_as_string().decode("utf-8")
        self.assertEqual(actual_content, content)

    def test_create_bucket_if_not_exists(self):
        self.__ems_storage_client.create_bucket_if_not_exists(TOOLKIT_CREATED_BUCKET, project=GCP_PROJECT_ID,
                                                              location="europe-west1")

        bucket = self.__storage_client.bucket(TOOLKIT_CREATED_BUCKET)
        self.assertTrue(bucket.exists())

    def test_create_bucket_if_not_exists_doesNothingIfExists(self):
        self.bucket.blob("create_bucket_test_blob.txt").upload_from_string("Test data")
        self.__ems_storage_client.create_bucket_if_not_exists(IT_TEST_BUCKET, project=GCP_PROJECT_ID,
                                                              location="europe-west1")

        bucket = self.__storage_client.bucket(IT_TEST_BUCKET)
        self.assertTrue(bucket.exists())
        self.bucket.blob("create_bucket_test_blob.txt").exists()

    def test_delete_bucket_if_exists_DeletesEmptyBucket(self):
        bucket_name = self.__generate_test_name("bucket")
        self.__ems_storage_client.create_bucket_if_not_exists(bucket_name, GCP_PROJECT_ID, "europe-west1")

        self.__ems_storage_client.delete_bucket_if_exists(bucket_name)

        bucket = self.__storage_client.bucket(bucket_name)
        self.assertFalse(bucket.exists())

    def test_delete_bucket_if_exists_DeletesNotExistingBucket(self):
        bucket_name = self.__generate_test_name("bucket")

        try:
            self.__ems_storage_client.delete_bucket_if_exists(bucket_name)
        except NotFound:
            self.fail("Should not throw exception if bucket is not exist")

    def test_delete_bucket_if_exists_DeletesNotEmptyBucket(self):
        bucket_name = self.__generate_test_name("bucket")
        self.__ems_storage_client.create_bucket_if_not_exists(bucket_name, GCP_PROJECT_ID, "europe-west1")
        self.__ems_storage_client.upload_from_string(bucket_name, "test.txt", "ok")

        self.__ems_storage_client.delete_bucket_if_exists(bucket_name, force=True)

        bucket = self.__storage_client.bucket(bucket_name)
        self.assertFalse(bucket.exists())

    def test_delete_blob(self):
        blob_name = "delete_blob_test_subject.txt"
        self.bucket.blob(blob_name).upload_from_string("foo")
        self.__ems_storage_client.delete_blob(IT_TEST_BUCKET, blob_name)
        self.assertFalse(self.bucket.blob(blob_name).exists())

    def test_create_notification_if_not_exists_noNotificationExistsCreatesIt(self):
        topic_name, bucket_name = self.__setup_notification_test_dependencies()

        self.__ems_storage_client.create_notification_if_not_exists(topic_name, bucket_name)

        notification_list = self.__storage_client.bucket(bucket_name).list_notifications()
        self.assertEqual(1, len(list(notification_list)))

        self.__cleanup([topic_name], bucket_name)

    def test_create_notification_if_not_exists_notificationExistsDoNotCreateIt(self):
        topic_name, bucket_name = self.__setup_notification_test_dependencies()

        self.__ems_storage_client.create_notification_if_not_exists(topic_name, bucket_name)
        self.__ems_storage_client.create_notification_if_not_exists(topic_name, bucket_name)

        notification_list = self.__storage_client.bucket(bucket_name).list_notifications()
        self.assertEqual(1, len(list(notification_list)))

        self.__cleanup([topic_name], bucket_name)

    def test_create_notification_if_not_exists_OtherNotificationExistsCreatesIt(self):
        topic_name, bucket_name = self.__setup_notification_test_dependencies()
        topic_name_other = topic_name + "_other"
        self.__ems_publisher_client.topic_create_if_not_exists(GCP_PROJECT_ID, topic_name_other)
        self.__ems_storage_client.create_notification_if_not_exists(topic_name_other, bucket_name)

        self.__ems_storage_client.create_notification_if_not_exists(topic_name, bucket_name)

        notification_list = self.__storage_client.bucket(bucket_name).list_notifications()
        self.assertEqual(2, len(list(notification_list)))

        self.__cleanup([topic_name, topic_name_other], bucket_name)

    def test_delete_notification_if_exists(self):
        topic_name, bucket_name = self.__setup_notification_test_dependencies()
        self.__ems_storage_client.create_notification_if_not_exists(topic_name, bucket_name)

        self.__ems_storage_client.delete_notification_if_exists(topic_name, bucket_name)

        notification_list = self.__storage_client.bucket(bucket_name).list_notifications()
        self.assertEqual(0, len(list(notification_list)))

        self.__cleanup([topic_name], bucket_name)

    def test_delete_notification_if_exists_WontFailWhenNotificationWasNotCreated(self):
        topic_name, bucket_name = self.__setup_notification_test_dependencies()

        self.__ems_storage_client.delete_notification_if_exists(topic_name, bucket_name)

        notification_list = self.__storage_client.bucket(bucket_name).list_notifications()
        self.assertEqual(0, len(list(notification_list)))

        self.__cleanup([topic_name], bucket_name)

    def test_delete_notification_if_exists_DeletesOnlyOneOfTheNotifications(self):
        topic_name, bucket_name = self.__setup_notification_test_dependencies()
        topic_name_other = topic_name + "_other"
        self.__ems_publisher_client.topic_create_if_not_exists(GCP_PROJECT_ID, topic_name_other)
        self.__ems_storage_client.create_notification_if_not_exists(topic_name, bucket_name)
        self.__ems_storage_client.create_notification_if_not_exists(topic_name_other, bucket_name)

        self.__ems_storage_client.delete_notification_if_exists(topic_name, bucket_name)

        notification_list = self.__storage_client.bucket(bucket_name).list_notifications()
        self.assertEqual(1, len(list(notification_list)))

        self.__cleanup([topic_name, topic_name_other], bucket_name)

    def __setup_notification_test_dependencies(self):
        topic_name = self.__generate_test_name("topic")
        bucket_name = self.__generate_test_name("bucket")
        self.__ems_publisher_client.topic_create_if_not_exists(GCP_PROJECT_ID, topic_name)
        self.__ems_storage_client.create_bucket_if_not_exists(bucket_name, GCP_PROJECT_ID, "europe-west1")
        return topic_name, bucket_name

    def __cleanup(self, topic_names, bucket_name):
        for topic_name in topic_names:
            self.__ems_publisher_client.delete_topic_if_exists(GCP_PROJECT_ID, topic_name)
        bucket = self.__storage_client.bucket(bucket_name)
        if bucket.exists():
            bucket.delete(force=True)

    @staticmethod
    def __generate_test_name(context: str):
        return "test_" + context + "_" + str(int(time.time()))



from google.cloud.pubsub_v1.subscriber.futures import StreamingPullFuture


class EmsStreamingFuture:

    def __init__(self, streaming_pull_future: StreamingPullFuture):
        self.__future = streaming_pull_future

    def result(self) -> str:
        return self.__future.result()

    def cancel(self):
        return self.__future.cancel()



from datetime import datetime
from typing import List, Union

from bigquery.job.config.ems_extract_job_config import EmsExtractJobConfig
from bigquery.job.ems_job import EmsJob
from bigquery.job.ems_job_state import EmsJobState


class EmsExtractJob(EmsJob):
    def __init__(self,
                 job_id: str,
                 table: str,
                 destination_uris: List[str],
                 job_config: EmsExtractJobConfig,
                 state: EmsJobState,
                 error_result: Union[dict, None],
                 created: datetime = None):
        super(EmsExtractJob, self).__init__(job_id, state, error_result, created)

        self.__job_config = job_config
        self.__table = table
        self.__destination_uris = destination_uris

    @property
    def table(self) -> str:
        return self.__table

    @property
    def job_config(self) -> EmsExtractJobConfig:
        return self.__job_config

    @property
    def destination_uris(self) -> List[str]:
        return self.__destination_uris



from enum import Enum

from bigquery.job.config.ems_job_config import EmsJobConfig


class DestinationFormat(Enum):
    CSV = "CSV"
    NEWLINE_DELIMITED_JSON = "NEWLINE_DELIMITED_JSON"
    AVRO = "AVRO"


class Compression(Enum):
    GZIP = "GZIP"
    DEFLATE = "DEFLATE"
    SNAPPY = "SNAPPY"
    NONE = "NONE"


class EmsExtractJobConfig(EmsJobConfig):

    def __init__(self,
                 compression: Compression = Compression.NONE,
                 destination_format: DestinationFormat = DestinationFormat.CSV,
                 field_delimiter: str = None,
                 print_header: bool = False,
                 *args, **kwargs):
        super(EmsExtractJobConfig, self).__init__(*args, **kwargs)
        self.__print_header = print_header
        self.__destination_format = destination_format
        self.__field_delimiter = field_delimiter
        self.__compression = compression

    @property
    def print_header(self):
        return self.__print_header

    @property
    def destination_format(self):
        return self.__destination_format

    @property
    def field_delimiter(self):
        return self.__field_delimiter

    @property
    def compression(self):
        return self.__compression



import time
from unittest import TestCase

from google.api_core.exceptions import NotFound, AlreadyExists
from google.cloud.pubsub_v1 import PublisherClient, SubscriberClient

from pubsub.ems_publisher_client import EmsPublisherClient
from tests.integration import GCP_PROJECT_ID


class ItEmsPublisherClient(TestCase):

    def setUp(self):
        self.__ems_publisher_client = EmsPublisherClient()
        self.__publisher_client = PublisherClient()

    def test_topic_create_if_not_exists_new_topic_creation(self):
        expected_topic_name = self.__generate_test_name("topic")
        expected_topic_path = self.__publisher_client.api.topic_path(GCP_PROJECT_ID, expected_topic_name)

        self.__ems_publisher_client.topic_create_if_not_exists(GCP_PROJECT_ID, expected_topic_name)

        try:
            topic = self.__publisher_client.api.get_topic(request={"topic": expected_topic_path})
        except NotFound:
            self.fail(f"Topic not created with name {expected_topic_name}")

        self.assertEqual(topic.name, expected_topic_path)

        self.__delete_topic(expected_topic_name)

    def test_topic_create_if_not_exists_topic_already_created(self):
        expected_topic_name = self.__generate_test_name("topic")

        self.__ems_publisher_client.topic_create_if_not_exists(GCP_PROJECT_ID, expected_topic_name)

        try:
            self.__ems_publisher_client.topic_create_if_not_exists(GCP_PROJECT_ID, expected_topic_name)
        except AlreadyExists:
            self.fail(f"Topic already exists but tried to recreate with name {expected_topic_name}")

        self.__delete_topic(expected_topic_name)

    def test_subscription_create(self):
        expected_topic_name = self.__generate_test_name("topic")
        expected_subscription_name = self.__generate_test_name("subscription")
        expected_subscription_list = ["projects/" + GCP_PROJECT_ID + "/subscriptions/" + expected_subscription_name]
        subscription_list = []

        self.__ems_publisher_client.topic_create_if_not_exists(GCP_PROJECT_ID, expected_topic_name)
        self.__ems_publisher_client.subscription_create(GCP_PROJECT_ID, expected_topic_name, expected_subscription_name)

        try:
            topic_path = self.__publisher_client.api.topic_path(GCP_PROJECT_ID, expected_topic_name)
            subscriptions = self.__publisher_client.api.list_topic_subscriptions(request={"topic": topic_path})

            for subscription in subscriptions:
                subscription_list.append(subscription)
        except NotFound:
            self.fail(
                f"Subscription not created with topic name {expected_topic_name}," +
                f"subscription name {expected_subscription_name}"
            )

        self.assertNotEqual(len(subscription_list), 0, "Subscription list is empty")
        self.assertEqual(subscription_list, expected_subscription_list, "Subscriptions not created")

        self.__delete_subscription(expected_subscription_name)
        self.__delete_topic(expected_topic_name)

    def test_delete_topic_if_exists(self):
        topic_name = self.__generate_test_name("topic")

        self.__ems_publisher_client.topic_create_if_not_exists(GCP_PROJECT_ID, topic_name)
        self.__ems_publisher_client.delete_topic_if_exists(GCP_PROJECT_ID, topic_name)
        try:
            self.__ems_publisher_client.delete_topic_if_exists(GCP_PROJECT_ID, topic_name)
        except NotFound:
            self.fail("delete_topic_if_exists raised NotFound error")

        topic_path = self.__publisher_client.api.topic_path(GCP_PROJECT_ID, topic_name)
        with self.assertRaises(NotFound):
            self.__publisher_client.api.get_topic(request={"topic": topic_path})

    def __delete_topic(self, topic_name):
        topic_path = self.__publisher_client.api.topic_path(GCP_PROJECT_ID, topic_name)
        self.__publisher_client.api.delete_topic(request={"topic": topic_path})

    @staticmethod
    def __delete_subscription(subscription_name: str):
        subscriber_client = SubscriberClient()
        subscription_path = subscriber_client.api.subscription_path(GCP_PROJECT_ID, subscription_name)
        subscriber_client.api.delete_subscription(request={"subscription": subscription_path})

    @staticmethod
    def __generate_test_name(context: str):
        return "test_" + context + "_" + str(int(time.time()))



from bigquery.job.config.ems_job_config import EmsJobConfig


class EmsLoadJobConfig(EmsJobConfig):
    
    def __init__(self, schema: dict, source_uri_template: str, skip_leading_rows: int = 0, *args, **kwargs):
        super(EmsLoadJobConfig, self).__init__(*args, **kwargs)
        self.__schema_json = schema
        self.__source_uri_template = source_uri_template
        self.__skip_leading_rows = skip_leading_rows
        self.__validate(self.destination_project_id)
        self.__validate(self.destination_dataset)
        self.__validate(self.destination_table)

    @property
    def source_uri_template(self):
        return self.__source_uri_template

    @property
    def schema(self):
        return self.__schema_json

    @property
    def skip_leading_rows(self):
        return self.__skip_leading_rows

    @staticmethod
    def __validate(value):
        if value is None or value.strip() == "":
            raise ValueError("Cannot be None or empty string!")



import datetime
from unittest import TestCase

from googleapiclient import discovery
from tenacity import retry, stop_after_delay, retry_if_result, wait_fixed

from cloudsql.ems_cloudsql_client import EmsCloudsqlClient
from cloudsql.ems_cloudsql_util import EmsCloudsqlUtil, TempBucketDescriptor
from storage.ems_storage_client import EmsStorageClient
from tests.integration import GCP_PROJECT_ID
from tests.integration import GCP_CLOUDSQL_INSTANCE_ID

DATABASE = "ems-gcp-toolkit-test"
DISCOVERY_SERVICE = discovery.build("sqladmin", "v1beta4", cache_discovery=False)
BUCKET_NAME = GCP_PROJECT_ID + "-gcp-toolkit-it"
IMPORT_USER = "postgres"
JOB_TIMEOUT_SECONDS = 30


# TODO  extract common test utils for
# (__get_table_content, get test bucket, __create_input_csv, __drop_and_create_table), retry

class ItEmsCloudsqlUtil(TestCase):

    def setUp(self):
        self.__storage_client = EmsStorageClient(GCP_PROJECT_ID)
        temp_bucket = TempBucketDescriptor(
            GCP_PROJECT_ID,
            GCP_CLOUDSQL_INSTANCE_ID + "-temp-bucket",
            "europe-west1"
        )
        self.__cloud_sql_client = EmsCloudsqlClient(GCP_PROJECT_ID,
                                                    GCP_CLOUDSQL_INSTANCE_ID)

        self.__util = EmsCloudsqlUtil(self.__cloud_sql_client, self.__storage_client, temp_bucket)
        self.__storage_client.create_bucket_if_not_exists(BUCKET_NAME, GCP_PROJECT_ID, "europe-west1")

    def test_run_sql(self):
        table_name = "test_table"
        suffix = str(int(datetime.datetime.utcnow().timestamp()))
        blob_name = f"export_{suffix}.csv"
        query = f'''DROP TABLE IF EXISTS {table_name};
                              CREATE TABLE {table_name} (id INTEGER PRIMARY KEY, name VARCHAR);
                              INSERT INTO {table_name} VALUES (3, 'old foo'), (4, 'old bar');'''

        self.__util.run_sql(DATABASE, query, IMPORT_USER, JOB_TIMEOUT_SECONDS)

        self.__export_table_to_csv(f"gs://{BUCKET_NAME}/{blob_name}", table_name)
        lines = self.__storage_client.download_lines(BUCKET_NAME, blob_name, num_lines=3)
        self.assertIn("3,old foo", lines)
        self.assertIn("4,old bar", lines)

    def __get_table_content(self, table_name):
        suffix = str(int(datetime.datetime.utcnow().timestamp()))
        blob_name = f"export_{suffix}.csv"
        self.__export_table_to_csv(f"gs://{BUCKET_NAME}/{blob_name}", table_name)
        content = self.__storage_client.download_content_as_string(BUCKET_NAME, blob_name)
        self.__storage_client.delete_blob(BUCKET_NAME, blob_name)
        return content

    def __create_input_csv(self, content):
        suffix = str(int(datetime.datetime.utcnow().timestamp()))
        blob_name = f"input_{suffix}.csv"
        self.__storage_client.upload_from_string(BUCKET_NAME, blob_name, content)

        return f"gs://{BUCKET_NAME}/{blob_name}"

    def __create_table_with_dumy_values(self, table_name):
        create_table_sql = f'''DROP TABLE IF EXISTS {table_name};
                                      CREATE TABLE {table_name} (id INTEGER PRIMARY KEY, name VARCHAR);
                                       INSERT INTO {table_name} VALUES (3, 'old foo'), (4, 'old bar');'''

        blob_name = f"create_test_table.sql"
        self.__storage_client.upload_from_string(BUCKET_NAME, blob_name, create_table_sql)

        blob_name = f"create_test_table.sql"
        request_body = {
            "importContext": {
                "kind": "sql#importContext",
                "fileType": "SQL",
                "uri": "gs://" + BUCKET_NAME + "/" + blob_name,
                "database": DATABASE,
                "importUser": IMPORT_USER
            }
        }
        request = DISCOVERY_SERVICE.instances().import_(project=GCP_PROJECT_ID,
                                                        instance=GCP_CLOUDSQL_INSTANCE_ID,
                                                        body=request_body)
        self.__wait_for_job_done(request.execute()["name"])

        self.__storage_client.delete_blob(BUCKET_NAME, blob_name)

    def __export_table_to_csv(self, export_uri, table_name):
        export_request_body = {
            "exportContext": {
                "kind": "sql#exportContext",
                "fileType": "CSV",
                "uri": export_uri,
                "databases": [
                    DATABASE
                ],
                "csvExportOptions": {
                    "selectQuery": f"select * from {table_name} "
                }
            }
        }

        request = DISCOVERY_SERVICE.instances().export(project=GCP_PROJECT_ID,
                                                       instance=GCP_CLOUDSQL_INSTANCE_ID,
                                                       body=export_request_body)
        response = request.execute()

        status = self.__wait_for_job_done(response["name"])
        assert "error" not in status, f"Status: {status}"

    @staticmethod
    @retry(stop=(stop_after_delay(JOB_TIMEOUT_SECONDS)),
           retry=(retry_if_result(lambda result: result["status"] != "DONE")),
           wait=wait_fixed(2))
    def __wait_for_job_done(ops_id):
        ops_request = DISCOVERY_SERVICE.operations().get(project=GCP_PROJECT_ID, operation=ops_id)
        ops_response = ops_request.execute()
        return ops_response



import logging
from builtins import BrokenPipeError

from googleapiclient import discovery
from googleapiclient.errors import HttpError
from tenacity import retry, stop_after_delay, retry_if_result, wait_fixed, retry_if_exception_type

LOGGER = logging.getLogger(__name__)


class EmsCloudsqlClient:
    EXECUTE_TIMEOUT = 600
    IMPORT_CSV_TIMEOUT = 600
    RELOAD_TABLE_TIMEOUT = 600
    CREATE_TMP_TABLE_TIMEOUT = 30

    def __init__(self, project_id: str, instance_id: str):
        self.__project_id = project_id

        self.__instance_id = instance_id
        self.__discovery_service = discovery.build("sqladmin", "v1beta4", cache_discovery=False)

    @property
    def project_id(self) -> str:
        return self.__project_id

    @property
    def instance_id(self) -> str:
        return self.__instance_id

    def import_csv_from_bucket(self, database: str, destination_table_name: str, source_csv_uri: str,
                               timeout_seconds: float) -> None:
        LOGGER.info("Importing CSV from %s to table %s in database %s",
                    source_csv_uri, destination_table_name, database)
        import_request_body = {
            "importContext": {
                "kind": "sql#importContext",
                "fileType": "CSV",
                "uri": source_csv_uri,
                "database": database,
                "csvImportOptions": {
                    "table": destination_table_name
                }
            }
        }
        request = self.__discovery_service.instances().import_(project=self.__project_id,
                                                               instance=self.__instance_id,
                                                               body=import_request_body)
        ops_id = EmsCloudsqlClient.execute_request(request)
        self.__wait_for_job_done(ops_id, timeout_seconds)

    def import_sql_from_bucket(self, database: str, source_sql_uri: str, timeout_seconds: float,
                               import_user: str) -> None:
        LOGGER.info("Importing SQL from %s to database %s", source_sql_uri, database)
        request_body = {
            "importContext": {
                "kind": "sql#importContext",
                "fileType": "SQL",
                "uri": source_sql_uri,
                "database": database,
                "importUser": import_user
            }
        }
        request = self.__discovery_service.instances().import_(project=self.__project_id,
                                                               instance=self.__instance_id,
                                                               body=request_body)
        ops_id = EmsCloudsqlClient.execute_request(request)
        self.__wait_for_job_done(ops_id, timeout_seconds)

    @staticmethod
    @retry(wait=wait_fixed(15),
           stop=stop_after_delay(EXECUTE_TIMEOUT),
           retry=retry_if_exception_type(HttpError))
    def execute_request(request) -> str:
        return request.execute()["name"]

    def __wait_for_job_done(self, ops_id: str, timeout_seconds: float) -> None:
        LOGGER.info("Waiting for job %s to be done", ops_id)

        @retry(wait=wait_fixed(1),
               stop=stop_after_delay(timeout_seconds),
               retry=(
                   retry_if_result(lambda result: result["status"] != "DONE")
                   | retry_if_exception_type(HttpError)
                   | retry_if_exception_type(BrokenPipeError)
               )
               )
        def __wait_for_job_done_helper() -> dict:
            ops_request = self.__discovery_service.operations().get(project=self.__project_id, operation=ops_id)
            ops_response = ops_request.execute()
            return ops_response

        status = __wait_for_job_done_helper()
        if "error" in status:
            raise EmsCloudsqlClientError(f"job failed with error status {status}")


class EmsCloudsqlClientError(Exception):
    pass



from datetime import datetime
from typing import Union

from bigquery.job.config.ems_query_job_config import EmsQueryJobConfig
from bigquery.job.ems_job import EmsJob
from bigquery.job.ems_job_state import EmsJobState


class EmsQueryJob(EmsJob):
    def __init__(self,
                 job_id: str,
                 query: str,
                 query_config: EmsQueryJobConfig,
                 state: EmsJobState,
                 error_result: Union[dict, None],
                 created: datetime = None):
        super(EmsQueryJob, self).__init__(job_id, state, error_result, created)

        self.__query = query
        self.__query_config = query_config

    @property
    def query_config(self) -> EmsQueryJobConfig:
        return self.__query_config

    @property
    def query(self) -> str:
        return self.__query



from datetime import datetime
from typing import Union

from bigquery.job.config.ems_load_job_config import EmsLoadJobConfig
from bigquery.job.ems_job import EmsJob
from bigquery.job.ems_job_state import EmsJobState


class EmsLoadJob(EmsJob):
    def __init__(self,
                 job_id: str,
                 load_config: EmsLoadJobConfig,
                 state: EmsJobState,
                 error_result: Union[dict, None],
                 created: datetime = None):
        super(EmsLoadJob, self).__init__(job_id, state, error_result, created)

        self.__load_config = load_config

    @property
    def load_config(self) -> EmsLoadJobConfig:
        return self.__load_config



from unittest import TestCase
from unittest.mock import patch, Mock

from storage.ems_storage_client import EmsStorageClient


class TestEmsStorageClient(TestCase):

    @patch("storage.ems_storage_client.storage")
    def test_download_lines_handlesLineEndingsProperly(self, storage_module):
        mock_client = Mock()
        mock_bucket = Mock()
        mock_blob = Mock()
        storage_module.Client.return_value = mock_client
        mock_client.bucket.return_value = mock_bucket
        mock_bucket.blob.return_value = mock_blob
        blob_content = "ok" + chr(456789)
        harmed_encoded_blob_content_with_partial_bytes_on_end = blob_content.encode("utf-8")[0:5]
        mock_blob.download_as_string.return_value = harmed_encoded_blob_content_with_partial_bytes_on_end

        ems_client = EmsStorageClient("some-project-id")
        result_lines = ems_client.download_lines("some-bucket-name", "some-blob", num_lines=1, chunk_size=5)

        self.assertEqual(result_lines[0], "ok")

    @patch("storage.ems_storage_client.storage")
    def test_is_blob_empty_should_return_true_for_empty_blob(self, storage_module):
        mock_client = Mock()
        mock_bucket = Mock()
        mock_blob = Mock()
        storage_module.Client.return_value = mock_client
        mock_client.get_bucket.return_value = mock_bucket
        mock_bucket.get_blob.return_value = mock_blob
        mock_blob.size = 0

        ems_client = EmsStorageClient("some-project-id")

        self.assertTrue(ems_client.is_blob_empty("bucket_name", "blob_name"))

    @patch("storage.ems_storage_client.storage")
    def test_is_blob_empty_should_return_false_for_non_empty_blob(self, storage_module):
        mock_client = Mock()
        mock_bucket = Mock()
        mock_blob = Mock()
        storage_module.Client.return_value = mock_client
        mock_client.get_bucket.return_value = mock_bucket
        mock_bucket.get_blob.return_value = mock_blob
        mock_blob.size = 1

        ems_client = EmsStorageClient("some-project-id")

        self.assertFalse(ems_client.is_blob_empty("bucket_name", "blob_name"))



import datetime
from unittest import TestCase

from google.cloud import storage
from google.cloud.exceptions import NotFound
from googleapiclient import discovery
from tenacity import retry, stop_after_delay, retry_if_result, wait_fixed

from cloudsql.ems_cloudsql_client import EmsCloudsqlClient
from tests.integration import GCP_PROJECT_ID
from tests.integration import GCP_CLOUDSQL_INSTANCE_ID


class ItEmsCloudsqlClient(TestCase):
    DATABASE = "ems-gcp-toolkit-test"
    DISCOVERY_SERVICE = discovery.build("sqladmin", "v1beta4", cache_discovery=False)
    BUCKET_NAME = GCP_PROJECT_ID + "-gcp-toolkit-it"
    IMPORT_USER = "postgres"
    JOB_TIMEOUT_SECONDS = 30

    def setUp(self):
        self.__storage_client = storage.Client(GCP_PROJECT_ID)
        self.__client = EmsCloudsqlClient(GCP_PROJECT_ID,
                                          GCP_CLOUDSQL_INSTANCE_ID)

    def test_import_csv_from_bucket_importsData(self):
        table_name = "testtable"
        self.__drop_and_create_table(table_name)

        csv_uri = self.__create_input_csv(f"1,alma\n")

        self.__client.import_csv_from_bucket(self.DATABASE, table_name, csv_uri, self.JOB_TIMEOUT_SECONDS)

        loaded_data = self.__get_table_content(table_name)
        self.assertEqual(loaded_data, f"1,alma\n")

    def test_import_sql_from_bucket_importsSql(self):
        table_name = "existing"
        content_to_load = f'''DROP TABLE IF EXISTS {table_name};
                              CREATE TABLE {table_name} (id INTEGER PRIMARY KEY, name VARCHAR);
                              INSERT INTO {table_name} VALUES (1, 'foo');'''
        bucket = self.__get_test_bucket(self.BUCKET_NAME)
        suffix = str(int(datetime.datetime.utcnow().timestamp()))
        blob_name = f"input_{suffix}.csv"
        blob = bucket.blob(blob_name)
        blob.upload_from_string(content_to_load)
        source_uri = f"gs://{self.BUCKET_NAME}/{blob_name}"

        self.__client.import_sql_from_bucket(self.DATABASE, source_uri, 30, self.IMPORT_USER)

        loaded_data = self.__get_table_content(table_name)
        self.assertEqual(loaded_data, "1,foo\n")

    def __get_test_bucket(self, bucket_name=BUCKET_NAME):

        try:
            bucket = self.__storage_client.get_bucket(bucket_name)
        except NotFound:
            bucket = self.__storage_client.bucket(bucket_name)
            bucket.location = "europe-west1"
            bucket.storage_class = "REGIONAL"
            bucket.create()
        return bucket

    # error cases:
    # blob does not exist
    # bucket does not exist
    # import was unsuccessful
    # delete temp blobs

    def __drop_and_create_table(self, table_name):
        create_table_sql = f'''DROP TABLE IF EXISTS {table_name};
                              CREATE TABLE {table_name} (id INTEGER PRIMARY KEY, name VARCHAR);'''

        blob_name = f"create_test_table.sql"
        self.__get_test_bucket().blob(blob_name).upload_from_string(create_table_sql)
        request_body = {
            "importContext": {
                "kind": "sql#importContext",
                "fileType": "SQL",
                "uri": "gs://" + self.BUCKET_NAME + "/" + blob_name,
                "database": self.DATABASE,
                "importUser": self.IMPORT_USER
            }
        }
        request = self.DISCOVERY_SERVICE.instances().import_(project=GCP_PROJECT_ID,
                                                             instance=GCP_CLOUDSQL_INSTANCE_ID,
                                                             body=request_body)
        self.__wait_for_job_done(request.execute()["name"])

        self.__get_test_bucket().blob(blob_name).delete()

    def __create_input_csv(self, content):
        bucket = self.__get_test_bucket(self.BUCKET_NAME)
        suffix = str(int(datetime.datetime.utcnow().timestamp()))
        blob_name = f"input_{suffix}.csv"
        blob = bucket.blob(blob_name)
        blob.upload_from_string(content)
        return f"gs://{self.BUCKET_NAME}/{blob_name}"

    def __get_table_content(self, table_name):
        suffix = str(int(datetime.datetime.utcnow().timestamp()))
        bucket = self.__get_test_bucket(self.BUCKET_NAME)
        blob_name = f"export_{suffix}.csv"
        self.__export_table_to_csv(f"gs://{self.BUCKET_NAME}/{blob_name}", table_name)
        content = bucket.blob(blob_name).download_as_string().decode("utf-8")
        bucket.delete_blob(blob_name)
        return content

    def __export_table_to_csv(self, export_uri, table_name):
        export_request_body = {
            "exportContext": {
                "kind": "sql#exportContext",
                "fileType": "CSV",
                "uri": export_uri,
                "databases": [
                    self.DATABASE
                ],
                "csvExportOptions": {
                    "selectQuery": f"select * from {table_name} "
                }
            }
        }

        request = self.DISCOVERY_SERVICE.instances().export(project=GCP_PROJECT_ID,
                                                            instance=GCP_CLOUDSQL_INSTANCE_ID,
                                                            body=export_request_body)
        response = request.execute()

        status = self.__wait_for_job_done(response["name"])
        assert "error" not in status, f"Status: {status}"

    @retry(stop=(stop_after_delay(JOB_TIMEOUT_SECONDS)),
           retry=(retry_if_result(lambda result: result["status"] != "DONE")),
           wait=wait_fixed(2))
    def __wait_for_job_done(self, ops_id):
        ops_request = self.DISCOVERY_SERVICE.operations().get(project=GCP_PROJECT_ID, operation=ops_id)
        ops_response = ops_request.execute()
        return ops_response

# delete temp table

# wait for bucket upload ????



from unittest import TestCase

from bigquery.job.config.ems_extract_job_config import EmsExtractJobConfig
from bigquery.job.ems_extract_job import EmsExtractJob
from bigquery.job.ems_job_state import EmsJobState

EXTRACT_JOB_CONFIG = EmsExtractJobConfig("comp", "format c:", "delimiter", False)


class TestEmsExtractJob(TestCase):

    def setUp(self):
        self.expected_error_result = {"some": "error", "happened": "here"}
        self.ems_extract_job = EmsExtractJob("test-job-id", "dummy-project-id.dummy-dataset.dummy-table",
                                             ["gs://some-non-existing-bucket-id/destination1"],
                                             EXTRACT_JOB_CONFIG,
                                             EmsJobState.DONE,
                                             self.expected_error_result)

    def test_state(self):
        self.assertEqual(self.ems_extract_job.state, EmsJobState.DONE)

    def test_job_id(self):
        self.assertEqual(self.ems_extract_job.job_id, "test-job-id")

    def test_is_failed(self):
        self.assertTrue(self.ems_extract_job.is_failed)

    def test_is_not_failed(self):
        not_failed_ems_load_job = EmsExtractJob("test-job-id", "dummy-project-id.dummy-dataset.dummy-table",
                                                ["gs://some-non-existing-bucket-id/destination1"], EXTRACT_JOB_CONFIG,
                                                EmsJobState.DONE, None)

        self.assertFalse(not_failed_ems_load_job.is_failed)

    def test_error_result(self):
        self.assertEqual(self.ems_extract_job.error_result, self.expected_error_result)



import logging
from typing import Callable, Iterator, List

from google.api_core.exceptions import AlreadyExists, NotFound
from google.cloud.pubsub_v1 import SubscriberClient
from google.cloud.pubsub_v1.subscriber.message import Message

from pubsub.ems_streaming_future import EmsStreamingFuture
from pubsub.ems_message import EmsMessage

LOGGER = logging.getLogger(__name__)


class EmsSubscriberClient:

    def __init__(self):
        self.__client = SubscriberClient()

    def subscribe(self, subscription: str, callback: Callable[[EmsMessage], None]) -> EmsStreamingFuture:
        def callback_wrapper(message: Message) -> None:
            callback(EmsMessage(message.ack_id, message.data, message.attributes))
            message.ack()

        future = self.__client.subscribe(
            subscription=subscription,
            callback=callback_wrapper
        )

        return EmsStreamingFuture(future)

    def pull(self,
             subscription: str,
             max_messages: int,
             return_immediately: bool = None) -> Iterator[EmsMessage]:
        messages = self.__client.api.pull(request={
            "subscription": subscription,
            "max_messages": max_messages,
            "return_immediately": return_immediately
        }).received_messages

        # pylint: disable=map-builtin-not-iterating
        return map(lambda msg: EmsMessage(msg.ack_id, msg.message.data, msg.message.attributes), messages)

    def acknowledge(self,
                    subscription: str,
                    ack_ids: List[str]) -> None:
        self.__client.api.acknowledge(request={"subscription": subscription, "ack_ids": ack_ids})

    def create_subscription_if_not_exists(self, project_id: str, topic_name: str, subscription_name: str):
        topic_path = self.__client.api.topic_path(project_id, topic_name)
        subscription_path = self.__client.api.subscription_path(project_id, subscription_name)
        try:
            self.__client.api.create_subscription(request={"name": subscription_path, "topic": topic_path})
            LOGGER.info("Subscription %s created for topic %s in project %s",
                        subscription_name, topic_name, project_id)
        except AlreadyExists:
            LOGGER.info("Subscription %s already exists for topic %s in project %s",
                        subscription_name, topic_name, project_id)

    def delete_subscription_if_exists(self, project_id: str, subscription_name: str):
        subscription_path = self.__client.api.subscription_path(project_id, subscription_name)
        try:
            self.__client.api.delete_subscription(request={"subscription": subscription_path})
            LOGGER.info("Subscription %s deleted in project %s", subscription_name, project_id)
        except NotFound:
            LOGGER.info("Subscription %s not found in project %s", subscription_name, project_id)



from setuptools import setup, find_packages

setup(
    name="ems-gcp-toolkit",
    version="0.2.1",
    packages=find_packages(exclude="tests"),
    url="https://github.com/emartech/ems-gcp-toolkit",
    license="MIT",
    author="Emarsys",
    author_email="",
    description="",
    install_requires=[
        "google-cloud-storage>=2,<3",
        "google-cloud-spanner>=3,<4",
        "google-cloud-pubsub>=2,<3",
        "google-cloud-bigquery>=3,<4",
        "google-cloud-core>=2,<3",
        "google-api-core>=2,<3",
        "googleapis-common-protos>=1.63.0,<2",
        "grpc-google-iam-v1==0.13.0"
    ]
)



from google.cloud.spanner import Client


class EmsSpannerClient:
    def __init__(self, project_id: str, instance_id, db_id):
        self.project_id = project_id
        self.__db = Client(project_id).instance(instance_id).database(db_id)

    def execute_sql(self, query: str):
        with self.__db.snapshot() as snapshot:
            results = snapshot.execute_sql(query)
            for row in results:
                yield row

    def execute_update(self, query: str):
        def run_query(transaction):
            transaction.execute_update(query)
        self.__db.run_in_transaction(run_query)


from enum import Enum

from bigquery.job.config.ems_job_config import EmsJobConfig, EmsJobPriority


class EmsTimePartitioningType(Enum):
    DAY = "DAY"


class EmsTimePartitioning:
    def __init__(self,
                 type_: EmsTimePartitioningType = None,
                 field: str = None,
                 expiration_ms: int = None,
                 require_partition_filter: bool = None):
        if type_ is None:
            self.__type = EmsTimePartitioningType.DAY
        else:
            self.__type = type_
        self.__field = field
        self.__expiration_ms = expiration_ms
        self.__require_partition_filter = require_partition_filter

    @property
    def type(self):
        return self.__type

    @property
    def field(self):
        return self.__field

    @property
    def expiration_ms(self):
        return self.__expiration_ms

    @property
    def require_partition_filter(self):
        return self.__require_partition_filter


class EmsQueryJobConfig(EmsJobConfig):

    def __init__(self,
                 priority: EmsJobPriority = EmsJobPriority.INTERACTIVE,
                 use_query_cache: bool = False,
                 time_partitioning: EmsTimePartitioning = None,
                 *args, **kwargs):
        super(EmsQueryJobConfig, self).__init__(*args, **kwargs)
        self.__priority = priority
        self.__use_query_cache = use_query_cache
        self.__time_partitioning = time_partitioning

    @property
    def priority(self):
        return self.__priority

    @property
    def use_query_cache(self):
        return self.__use_query_cache

    @property
    def time_partitioning(self):
        return self.__time_partitioning



from unittest import TestCase

from bigquery.job.config.ems_query_job_config import EmsQueryJobConfig
from bigquery.job.ems_job_state import EmsJobState
from bigquery.job.ems_query_job import EmsQueryJob


class TestEmsQueryJob(TestCase):


    def setUp(self):
        self.query_config = EmsQueryJobConfig()
        self.expected_error_result = {"some": "error", "happened": "here"}
        self.ems_query_job = EmsQueryJob("test-job-id", "query", self.query_config, EmsJobState.DONE,  self.expected_error_result)

    def test_state(self):
        self.assertEqual(self.ems_query_job.state, EmsJobState.DONE)

    def test_job_id(self):
        self.assertEqual(self.ems_query_job.job_id, "test-job-id")

    def test_query(self):
        self.assertEqual(self.ems_query_job.query, "query")

    def test_is_failed(self):
        self.assertTrue(self.ems_query_job.is_failed)

    def test_is_not_failed(self):
        not_failed_ems_query_job = EmsQueryJob("test-job-id", "query", self.query_config, EmsJobState.DONE, None)

        self.assertFalse(not_failed_ems_query_job.is_failed)

    def test_error_result(self):
        self.assertEqual(self.ems_query_job.error_result,  self.expected_error_result)



from unittest import TestCase

from bigquery.job.config.ems_job_config import EmsJobPriority, EmsWriteDisposition, EmsCreateDisposition

from bigquery.job.config.ems_query_job_config import EmsQueryJobConfig


class TestEmsQueryJobConfig(TestCase):
    def setUp(self):
        self.ems_query_job_config = EmsQueryJobConfig(priority=EmsJobPriority.INTERACTIVE,
                                                      destination_project_id="test_project",
                                                      destination_dataset="test_dataset",
                                                      destination_table="test_table",
                                                      create_disposition=EmsCreateDisposition.CREATE_IF_NEEDED,
                                                      write_disposition=EmsWriteDisposition.WRITE_APPEND,
                                                      use_query_cache=True,
                                                      labels={"function_name": "blabla"})

    def test_destination_project_id(self):
        self.assertEqual(self.ems_query_job_config.destination_project_id, "test_project")

    def test_destination_dataset(self):
        self.assertEqual(self.ems_query_job_config.destination_dataset, "test_dataset")

    def test_create_disposition(self):
        self.assertEqual(self.ems_query_job_config.create_disposition, EmsCreateDisposition.CREATE_IF_NEEDED)

    def test_write_disposition(self):
        self.assertEqual(self.ems_query_job_config.write_disposition, EmsWriteDisposition.WRITE_APPEND)

    def test_destination_table(self):
        self.assertEqual(self.ems_query_job_config.destination_table, "test_table")

    def test_priority(self):
        self.assertEqual(self.ems_query_job_config.priority, EmsJobPriority.INTERACTIVE)

    def test_use_query_cache(self):
        self.assertEqual(self.ems_query_job_config.use_query_cache, True)

    def test_labels(self):
        self.assertEqual(self.ems_query_job_config.labels, {"function_name": "blabla"})
